/*
* will parse a jwm config file to find keyboard shortcuts,
* heavily inspired by the application jwmConfigMgr by thoughtjourney.
* 
* compile anr run with:
* valac -X -Os --pkg libxml-2.0 --pkg gtk+-2.0 jwmshortcut.vala
* strip --strip-unneeded jwmshortcut
* ./jwmshortcut 
*/
using Gtk;
using Xml;

//an int delegate type (used to return command exit status...)
public delegate int IntType();

/*******************************************************************************
* 
* an horizontal box which gather a label and an entry field
* - just for GUI construction convenience.
* 
*******************************************************************************/
public class HBoxLabelledEntry : Gtk.HBox {

   public Gtk.Entry entry;

   public HBoxLabelledEntry(string label, string defaultentry){
      this.set_homogeneous(false);
      this.set_spacing(0);
      entry=new Gtk.Entry();
      entry.set_text(defaultentry);
      this.pack_end (entry, false, true, 0);
      this.pack_end (new Gtk.Label(label), false, true, 0);
   }
}

/*******************************************************************************
*
* Class to handle xml files:
* get an xml file path, parse it, and keep it ready to be read/modified 
* with help of xpaths to get to the desired nodes
* 
*******************************************************************************/
public class XmlFile : GLib.Object {
   
   private string xml_file;            // the file path
   private Xml.Doc* xml_doc;           // the main xml_doc
   private Xml.Doc* previous_xml_doc;  // a copy in case changes failed
   private Xml.XPathContext ctx;    
   
   
   public XmlFile(string path){
      
      this.xml_file = path;
      this.xml_doc=Xml.Parser.parse_file ( this.xml_file );
      
      if (this.xml_doc == null) 
         // Error: parsing failed
         return;  
      
      this.previous_xml_doc=this.xml_doc->copy(1);
      this.ctx= new Xml.XPathContext( xml_doc );   
   }
   
   ~XmlFile( ) {
      if ( this.xml_doc != null )
         delete this.xml_doc ;
       if ( this.previous_xml_doc != null )
         delete this.previous_xml_doc ;
   }
   
   // Return in a bi-dimensional array, the node's content [first position [0]] 
   // and all requested attributes [following positions [1-...]]
   // for node fitting the xpath description ... to be improved
   public void node_data (string xpath, out string[]? content, string[]? attr, out string[,]? attr_value){
      
      Xml.XPathObject* mainobjs=ctx.eval_expression( xpath );
      
      if ( mainobjs->nodesetval != null ){
         int node_nb=mainobjs->nodesetval->length();
         
         if (content != null){
            content=new string[node_nb];
            for(int i=0; i< node_nb; i++)
               content[i]=mainobjs->nodesetval->item(i)->get_content();           
         }
         
         if ( attr != null && attr_value != null ){
            attr_value=new string[attr.length,node_nb];
            for(int i=0; i< node_nb; i++)
               for(int a=0; a< attr.length; a++)
                  attr_value[a,i]=mainobjs->nodesetval->item(i)->get_prop(attr[a]);
         }
      }
      else{
         content=null;
         attr_value=null;
      }
   }
   
   // Remove node(s) fitting the given xpath from file
   public void node_remove(string xpath){
      Xml.XPathObject* mainobjs=ctx.eval_expression( xpath );
      
      if ( mainobjs->nodesetval != null )
         for(int i=0; i< mainobjs->nodesetval->length(); i++)
            mainobjs->nodesetval->item(i)->unlink();
   }
   
   // Modify some field for node(s) of the given xpath:
   public void node_modify(string xpath, string? content=null, string[,]? attr=null){
      Xml.XPathObject* mainobjs=ctx.eval_expression( xpath );
      
      if ( mainobjs->nodesetval != null ) 
         // then xpath is found
         for(int i=0; i< mainobjs->nodesetval->length(); i++)
         {
            if (content != null) 
               // then modify node's content
               mainobjs->nodesetval->item(i)->set_content(content);

            if (attr != null) 
               // then modify attributes from the string array
               for(int a=0; a<attr.length[1]; a++)
                  mainobjs->nodesetval->item(i)->set_prop(attr[0,a],attr[1,a]);
         }
   }
   
   public void node_addchild(string parent_xpath, string name, string? content=null, string[,]? attr=null){
      Xml.XPathObject* mainobjs=ctx.eval_expression( parent_xpath );
      
      // parent_xpath must be valid and unique 
      if ( mainobjs->nodesetval != null && mainobjs->nodesetval->length() == 1){
         mainobjs->nodesetval->item(0)->new_text_child(null,name,content);
         if (attr != null) 
            // then modify attributes from the string array
            for(int a=0; a<attr.length[1]; a++)
               mainobjs->nodesetval->item(0)->get_last_child()->new_prop(attr[0,a],attr[1,a]);
      }
   }
   
   // Try to print new shortcut to config file, cancel operation if new config is corrupt.   
   // open the config file and write the current xml_doc to it
   public void save_file (IntType? check_validity=null, IntType? post_action=null) {

      this.xml_doc->save_format_file(this.xml_file,this.xml_doc,1);
      
      if (check_validity != null && check_validity() != 0 ){
         // Delegate test function to check file's validity returned an error
         // so cancel changes: use the previous (clean) configuration file
         this.xml_doc=this.previous_xml_doc->copy(1);
         this.ctx= new Xml.XPathContext( xml_doc );
         save_file();
      }
      else{ 
         // Changes are valid, save a copy of this doc and do some final tasks
         this.previous_xml_doc=this.xml_doc->copy(1);
         // Perform any post changes action
         if (post_action != null) post_action();   
     }
   }
}

/*******************************************************************************
* 
* The main Shortcuts window, storing a Treeview to display the list of shortcuts
* and a pointer on a XmlFile to handle changes on the JWM config file
*  
*****************************************************************************/

public class Shortcuts : Gtk.Window {
 
   private Gtk.TreeView view;
   private XmlFile* jwmrc;
   
   private string webpage="'http://www.joewing.net/programs/jwm/config.shtml#keys'";
   private string shortcuts_info_text="""---
"mask" specifies the modifier keys that must be held to enable this action and "key" specifies the key that is to be pressed. 
---
The mask section is optional. It can contain any number of the following values: 

   A - Alt
   C - Control
   S - Shift
   H - Hyper
   M - Meta
   P - Super
---
"Action" is the action performed when this combination is pressed. 
---
See Help for more details
---
To make your changes effective, JWM may be restarted. 
---""";
   
   public Shortcuts (XmlFile* jwmrc) {
      
      this.jwmrc=jwmrc;
      
      // Buttons:
      var button_add=new Gtk.Button.from_stock(Gtk.STOCK_ADD);
      var button_remove=new Gtk.Button.from_stock(Gtk.STOCK_REMOVE);
      var button_edit=new Gtk.Button.from_stock(Gtk.STOCK_EDIT);
      var button_info=new Gtk.Button.from_stock(Gtk.STOCK_INFO);
      var button_help=new Gtk.Button.from_stock(Gtk.STOCK_HELP);
      var button_quit=new Gtk.Button.from_stock(Gtk.STOCK_QUIT);

      var vbox =new Gtk.VBox(true,5);
      vbox.add(new Gtk.Label(""));
      vbox.add(button_remove);
      vbox.add(button_add);
      vbox.add(button_edit);
      vbox.add(new Gtk.Label(""));
      vbox.add(button_info);
      vbox.add(button_help);
      vbox.add(button_quit);
      
      // Buttons'signals
      button_remove.clicked.connect (_remove);
      button_add.clicked.connect ((s)=>{ _editor("add"); });
      button_edit.clicked.connect ((s)=>{ _editor("edit"); });
      button_info.clicked.connect ((s)=>{
         _msg(shortcuts_info_text, "Shortcuts: How-to", Gtk.STOCK_INFO, Gtk.MessageType.INFO, this);
         });      
      button_help.clicked.connect ( (s)=>{_www(webpage);});
      button_quit.clicked.connect (Gtk.main_quit);      
        
      // Scrolled treeview
      this.view = new Gtk.TreeView ();
      this.view.set_rules_hint(true);
      _setup();
      _refresh();
      
      var scroll = new Gtk.ScrolledWindow (null, null);
      scroll.set_shadow_type (Gtk.ShadowType.ETCHED_IN);
      scroll.set_policy (Gtk.PolicyType.NEVER, Gtk.PolicyType.AUTOMATIC);
      scroll.add (this.view);        
      
      // Gather everything in the main window and show it
      var hbox=new Gtk.HBox(false,5);
      hbox.add(scroll);
      hbox.add(vbox);
      
      this.title = "Keyboard Shortcuts";
      this.set_focus(button_quit);
      this.border_width = 5;
      this.set_icon(this.render_icon (Gtk.STOCK_UNDERLINE, Gtk.IconSize.MENU, null));
      this.add(hbox);
      this.destroy.connect (Gtk.main_quit);
   }
 
   // Format column of the treeview
   private void _setup() {
      view.set_model (  new Gtk.ListStore (4, typeof (int), typeof (string),
                                       typeof (string), typeof (string))   );
      view.insert_column_with_attributes (-1, "Mask", new Gtk.CellRendererText (), "text", 1, null);
      view.insert_column_with_attributes (-1, "Key", new Gtk.CellRendererText (), "text", 2, null);
      view.insert_column_with_attributes (-1, "Action", new Gtk.CellRendererText (), "text", 3, null);
   }
   
   // Find shortcuts in configfile and refresh the treeview list
   private void _refresh() {
      Gtk.TreeIter iter;
      Gtk.ListStore listmodel= (Gtk.ListStore) this.view.model;
      listmodel.clear();
      
      string[] _content={};
      string[,] _attr_value={{}};
      this.jwmrc->node_data("/JWM/Key", out _content, {"mask","key"}, out _attr_value);
      
      if ( _content != null )
         for(int i=0; i< _content.length; i++)
         {
            listmodel.append (out iter);
            listmodel.set (iter, 0, i+1, 1, _attr_value[0,i], 2, _attr_value[1,i], 3, _content[i], -1);
         }
   }
   
   // Remove the selected shortcut from list/file
   private void _remove () {
      this.jwmrc->node_remove(add_index2string("/JWM/Key",_selectionindex()));
      _update();
   }
   
   // Update everything (config file and treeview list)
   private void _update (){
      // Check the validity of a jwm configuration file. Show error if it fails.
      // else restart jwm from the command line
      this.jwmrc->save_file(
         ()=>{return _sync_command("jwm -p",this);},
         ()=>{return _sync_command("jwm -restart",this);}
         );
      _refresh();
   }
   
   
   // Get the index of the current selection 
   // which is stored inside an hidden first column
   private int _selectionindex () {
      int i=-1;      // the index
      Gtk.TreeIter iter;
      Gtk.TreeSelection selection = this.view.get_selection();
      
      if ( selection.get_selected_rows(null).length() == 0 ) 
         return i;
      
      selection.get_selected(null, out iter);
      this.view.model.get( iter ,0 ,out i);
      return i;
   }
   
   // Launch a new window with entry for adding/modifying shortcut
   // case edit: editor with the current selection informations
   // case add: shortcut editor for a new shortcut input (default fields)
   private void _editor (string action_type) {
      
      // ShortcutEditor SE;
      string xpath_text="";
      string[] _content={"Action"};
      string[,] _attr_value={{"Mask"},{"Key"}};
      
      // Get the fields from the current selection in case of a modification
      if( action_type == "edit" ){
         xpath_text=add_index2string("/JWM/Key",_selectionindex());
         this.jwmrc->node_data(xpath_text, out _content, {"mask","key"}, out _attr_value);
         
         // mask is optional, if null, use empty string, else parse it
         _attr_value[0,0]=(_attr_value[0,0]!=null)?_attr_value[0,0]:"";  
      }
      
      // Build one HBoxLabelledEntry per field:
      var maskbox=new HBoxLabelledEntry("Mask (optional): ", _attr_value[0,0]);
      var keybox=new HBoxLabelledEntry("Key: ", _attr_value[1,0]);
      var actionbox=new HBoxLabelledEntry("Action: ", _content[0]);
      
      // New modal Dialog SE (for Shortcut Editor):
      var SE=new Gtk.Dialog.with_buttons("Shortcut Editor", this,
         Gtk.DialogFlags.MODAL | Gtk.DialogFlags.NO_SEPARATOR | Gtk.DialogFlags.DESTROY_WITH_PARENT, 
         Gtk.STOCK_HELP, Gtk.ResponseType.HELP, 
         Gtk.STOCK_OK, Gtk.ResponseType.OK);
      SE.set_focus(SE.add_button(Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL));

      SE.vbox.add (maskbox);
      SE.vbox.add (keybox);
      SE.vbox.add (actionbox);
      SE.set_icon(SE.render_icon (Gtk.STOCK_UNDERLINE, Gtk.IconSize.MENU, null));
      SE.vbox.spacing = 5;
      SE.border_width = 5;
      
      SE.response.connect ((s, response_id)=> {
           switch (response_id) {
           case Gtk.ResponseType.HELP:
               _www(webpage);
               break;              
           case Gtk.ResponseType.OK:
               
               // Read current values (stored in entry widgets)
               string mask_text = _parse_mask(maskbox.entry.get_text());
               string key_text = keybox.entry.get_text();
               string action_text= actionbox.entry.get_text();
               
               SE.destroy ();
               
               // Request xml and config file modification:
               if( action_type == "add" ) 
                  // then we create the shortcut from scratch:
                  this.jwmrc->node_addchild("/JWM","Key",action_text, 
                     {{"mask","key"},{mask_text,key_text}} );
               else                       
                  // then this is a modification:
                  this.jwmrc->node_modify(xpath_text, action_text, 
                     {{"mask","key"},{mask_text,key_text}} );
               
               _update();
               
               break;                  
            default:
               SE.destroy ();
               break;
           }
         });
         
      SE.show_all();
   }
   
   // Perform few substitution in mask entry to allow different kind of input
   string _parse_mask(string mask_text){
      
      string tmp=mask_text.up()
      .replace("ALT","1")
      .replace("CTRL","2")
      .replace("HYPER","3")
      .replace("META","4")
      .replace("SHIFT","5")
      .replace("SUPER","6")
      .replace("A","1")
      .replace("C","2")
      .replace("H","3")
      .replace("M","4")
      .replace("S","5")
      .replace("P","6");
      
      tmp.canon("123456",' ');
      
      tmp=tmp.replace(" ","")
      .replace("1","A")
      .replace("2","C")
      .replace("3","H")
      .replace("4","M")
      .replace("5","S")
      .replace("6","P");
      
      return tmp;
   }
}

/*******************************************************************************
* End of classes definition
*******************************************************************************/

// Function calling a "command" synchronously, print messages
// and return status
public int _sync_command(string command, Gtk.Window? parent=null){
   string? std_output="";
   string? std_error="";
   string? process_error="";
   
   try{ 
      GLib.Process.spawn_command_line_sync(command, out std_output, out std_error);
   }
   catch( GLib.SpawnError e){
      process_error="Error: Failed to execute the command: ".concat(command);
   }
   
   // Print errors:
   if( std_output != "" ) stdout.printf("standard output: %s", std_output);
   if( std_error != "" ) stdout.printf("standard error: %s", std_error);
   if( process_error != "" ) stdout.printf("Process error: %s", process_error);
   
   if ( std_error == "" && process_error == "")
      return 0;
   else{
      _msg("Cannot perform this request because it would corrupt the JWM configuration file. Read the help and the message bellow and you may find why.\n\nError details:".concat(std_output,"\n",std_error), 
      "Error", Gtk.STOCK_DIALOG_WARNING,Gtk.MessageType.ERROR, parent);
      return 1;
   }
}

// Concatenate a string and an int to "string[int]"
string add_index2string (string s, int i) {
   return s.concat("[", i.to_string(), "]", null);
}

// Standalone message dialog:
void _msg(string _text, string _title, string stock_id,
         Gtk.MessageType message_type,Gtk.Window? parent=null){
   var msg=new Gtk.MessageDialog (parent, Gtk.DialogFlags.DESTROY_WITH_PARENT, 
   message_type, Gtk.ButtonsType.CLOSE, _text);
   msg.title = _title;
   msg.set_icon(msg.render_icon (stock_id, Gtk.IconSize.MENU, null));
   msg.show_all();
   msg.response.connect ((s)=>{msg.destroy ();});
}

// Open the specified webpage
void _www (string webpage) {
   try{ 
      GLib.Process.spawn_command_line_async("defaulthtmlviewer ".concat(webpage) ); 
   }
   catch(GLib.SpawnError e){
      stdout.printf("Error: cannot open this web site with your default html viewer: %s",webpage);
   }   
}

/*******************************************************************************
* main:
*******************************************************************************/

public static void main(string[] args) {

   Gtk.init (ref args);
   Parser.init();
   
   // Parse the config file passed as first argument and show shorcuts list
   string jwmrc_path=( args[1]==null )?"/root/.jwm/jwmrc-personal":args[1];
   XmlFile* jwmrc=new XmlFile(jwmrc_path);
   
   var shortcuts_list=new Shortcuts(jwmrc);
   shortcuts_list.show_all();
   Gtk.main ();
   
   Parser.cleanup();
   delete jwmrc;
}